/* Copyright (C) 2005-2018 J.F.Dockes
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include "cdbrowser.h"

#include <string>

#include <QSettings>
#ifdef USING_WEBENGINE
#include <QLoggingCategory>
#endif
#include <libupnpp/upnpavutils.hxx>

#include "HelperStructs/CSettingsStorage.h"
#include "HelperStructs/Style.h"
#include "upadapt/upputils.h"
#include "utils/utf8iter.h"

using namespace UPnPClient;
using namespace UPnPP;

// Notes for WebEngine
// - All links must begin with http:// for acceptNavigationRequest to be
//   called. 
// - The links passed to acceptNav.. have the host part 
//   lowercased -> we change S0 to http://h/S0, not http://S0

// The const part + the js we need for webengine. This is logically
// const after initialization by the first constructor.
static QString const_html_top;

// The const part + the js + changeable style, computed as needed.
static QString html_top;

// Initial post-header content of the servers page. The servers list
// will be inserted in the body using Javascript
static const QString empty_server_page_bot(
    "</head><body>"
    "<h2 id='cdstitle'>Content Directory Services</h2>"
    "</body></html>"
    );

// Content while we look for servers. 
static const QString init_server_page_bot(
    "</head><body>"
    "<h2 id='cdstitle'>Content Directory Services</h2>"
    "<p>&nbsp;</p><p>Looking for servers...</p>"
    "</body></html>"
    );


// Once initialization of the constant part at the top of pages: read
// the javascript files.
static void init_const_html_top()
{
    const_html_top = u8s2qs(
        "<html><head><meta http-equiv='content-type' content='text/html; charset=utf-8'>");
#ifdef USING_WEBENGINE
    QLoggingCategory("js").setEnabled(QtDebugMsg, true);
    QString jsfn = Helper::getCSSPath() + "containerscript.js";
    QString wcfn = Helper::getCSSPath() + "qwebchannel.js";
    QString js = "<script type=\"text/javascript\">\n";
    js += QString::fromUtf8(Helper::readFileToByteArray(jsfn));
    js += "\n";
    js += QString::fromUtf8(Helper::readFileToByteArray(wcfn));
    js += "</script>\n";
    const_html_top += js;
#endif
}

void CDBrowser::init_HTML()
{
    if (const_html_top.isEmpty()) {
        init_const_html_top();
    }
    setStyleSheet(CSettingsStorage::getInstance()->getPlayerStyle(), false);
}

// Complete the top of pages by reading the CSS from our
// configuration, possibly fixing it for the current scale value, and
// appending it to const_html_top
void CDBrowser::setStyleSheet(bool dark, bool redisplay)
{
    QString cssfn = Helper::getCSSPath() + "cdbrowser.css";
    QString cssdata = QString::fromUtf8(Helper::readFileToByteArray(cssfn));

    float multiplier = QSettings().value("wholeuiscale").toFloat();
    if (multiplier == 0)
        multiplier = 1.0;

    // On some platforms (apple) we need to adjust the html font point
    // size because 10pt in the HTML part is bigger than 10pt in the GUI???
#ifdef __APPLE__
    float cssPointAdjust = 1.5;
#else
    float cssPointAdjust = 1.0;
#endif
    QSettings settings;
    if (settings.contains("htmlfontscaledownfactor")) {
        cssPointAdjust = settings.value("htmlfontscaledownfactor").toFloat();
    }
    multiplier /= cssPointAdjust;
    
    cssdata = u8s2qs(Style::scale_fonts(qs2utf8s(cssdata), multiplier));
    if (dark) {
        cssfn = Helper::getCSSPath() + "dark.css";
    } else {
        cssfn = Helper::getCSSPath() + "standard.css";
    }
    cssdata +=  Helper::readFileToByteArray(cssfn);

    html_top = const_html_top;
    html_top += "<style>\n";
    html_top += cssdata;
    html_top += "</style>\n";
    if (redisplay) {
        refresh();
    }
}

// With message "looking for servers"
QString CDBrowser::initialServersPage()
{
    return QString(html_top + init_server_page_bot);
}

// To be filled with the current servers list
QString CDBrowser::emptyServersPage()
{
    return QString(html_top + empty_server_page_bot);
}

QString CDBrowser::DSToHtml(unsigned int idx, const UPnPDeviceDesc& dev)
{
    QString out;
    out += QString("<p class='cdserver' cdsid='%1'>").arg(idx);
    out += QString("<a href='http://h/S%1'>").arg(idx);
    out += QString::fromUtf8(dev.friendlyName.c_str());
    out += QString("</a></p>");
    return out;
}

QString CDBrowser::imgtag(const std::string& imgurl, int height)
{
    QString url = imgurl.empty() ? m_formatparams.genericAlbumImageURI : u8s2qs(imgurl);
    // Crop the image without changing the aspect ratio
    // https://cloudinary.com/guides/automatic-image-cropping/5-ways-to-crop-images-in-html-css
    // Original approach: resize, lose aspect ratio
    //  return QString("<img height='%1' width='%1' src='%2' />").arg(height).arg(url);
    // Automatic centering (default object-position: 50% 50%:
    return QString("<img style='object-fit: cover; "
                   "height: %1; width: %1' src='%2' />").arg(height).arg(url);
}

static QString mkctattrs(const std::string& id, unsigned int idx, bool isPlaylist)
{
    return QString("class='container' objid='%1' objidx='%2' ispl='%3'").
        arg(u8s2qs(id)).arg(idx).arg(isPlaylist);
}

QString CDBrowser::CTToHtml(unsigned int idx, const UPnPDirObject& e, const QString& id)
{
    bool isPlaylist = !e.getprop("upnp:class").compare(plContainerClass);
    QString ctattrs = mkctattrs(e.m_id, idx, isPlaylist);
    
    std::string albumArtURI;
    if (m_formatparams.albumsAsCovers || m_formatparams.coverWithAlbum) {
        e.getprop("upnp:albumArtURI", albumArtURI);
    }
    std::string artist;
    if (m_formatparams.artistWithAlbum) {
        e.getprop("upnp:artist", artist);
        if (!artist.empty()) {
            utf8truncate(artist, m_formatparams.artistWithAlbumBytes,
                         UTF8T_ATWORD|UTF8T_ELLIPSIS, "\xE2\x80\xA6",
                         " \t\n\t,;.");
        }
    }
    
    QString title = u8s2qs(Helper::escapeHtml(e.m_title));
    QString idattr;
    if (!id.isEmpty())
        idattr = QString("id='%1'").arg(id);
    if (m_formatparams.albumsAsCovers) {
        int imgw = m_formatparams.albumsAsCoversWidth * m_formatparams.scalemultiplier;

        // Album title row. div with width limited to the cover image size.
        auto ttrow = QString(
            "<tr %1><td><div class='ct_albcovertt' style='width:%2px;'>"
            "<a class='ct_title' %4 href='http://h/C%3'>%5</a>"
            "</div></td></tr>")
            .arg(ctattrs).arg(imgw).arg(idx).arg(idattr).arg(title);

        // Artist row. Same.
        QString artrow = QString(
            "<tr %1><td><div class='ct_albcoverart' style='width:%2px;'>%3</div></td></tr>")
            .arg(ctattrs).arg(imgw).arg(u8s2qs(Helper::escapeHtml(artist)));
        
        // Whole block: table with 3 rows: image, title, artist
        QString thumbhtml = QString(
            "<table class='ct_albcovertable'>"
            "<tr %1><td><a href='http://h/C%2'>%3</a></td></tr>")
            .arg(ctattrs).arg(idx).arg(imgtag(albumArtURI, imgw)) +
            ttrow +
            artrow +
            "</table>";
        return thumbhtml;

    } else {
        // Displaying albums as lists, like tracks
        QString out = QString("<tr %1>").arg(ctattrs);

        if (m_formatparams.coverWithAlbum) {
            int imgw = m_formatparams.coverWithAlbumWidth * m_formatparams.scalemultiplier;
            out += QString("<td width='%1'><a class='ct_title' href='http://h/C%2' tabindex='-1'>%3</a></td>")
                .arg(imgw+2).arg(idx).arg(imgtag(albumArtURI, imgw));
        } else {
            out += QString("<td></td>");
        }

        out += QString("<td><a class='ct_title' %1 href='http://h/C%2'>%3</a></td>")
            .arg(idattr).arg(idx).arg(title);

        if (m_formatparams.artistWithAlbum) {
            out += QString("<td class='ct_artist'>%1</td>")
                .arg(u8s2qs(Helper::escapeHtml(artist)));
        }

        out += "</tr>";
        return out;
    }
}

/** 
 * Show item as table row with 5 cells: tno, artist, title, album, duration
 * @arg idx index in entries array
 * @arg e array entry
*/
QString CDBrowser::ItemToHtml(unsigned int idx, const UPnPDirObject& e,
                              int maxartlen)
{
    QString out;
    std::string val;

    out = QString("<tr class='item' objid='%1' objidx='%2'>").
        arg(e.m_id.c_str()).arg(idx);

    e.getprop("upnp:originalTrackNumber", val);
    out += QString("<td class='tk_tracknum'>");
    if (val.size() == 0) {
        out += "&nbsp;";
    } else {
        out += QString::fromUtf8(Helper::escapeHtml(val).c_str());
    }
    out += "</td>";

    out += "<td class='tk_title'>";
    if (e.m_title.size() == 0) {
        out += "&nbsp;";
    } else {
        out += QString("<a href='http://h/I%1'>").arg(idx);
        out += QString::fromUtf8(Helper::escapeHtml(e.m_title).c_str());
        out += "</a>";
    }
    out += "</td>";

    val.clear();
    e.getprop("upnp:artist", val);
    out += "<td class='tk_artist'>";
    if (maxartlen > 0 && int(val.size()) > maxartlen) {
        int len = maxartlen-3 >= 0 ? maxartlen-3 : 0;
        val = val.substr(0,len) + "...";
    }
    if (val.size() == 0) {
        out += "&nbsp;";
    } else {
        out += QString::fromUtf8(Helper::escapeHtml(val).c_str());
    }
    out += "</td>";
    
    val.clear();
    e.getprop("upnp:album", val);
    out += "<td class='tk_album'>";
    out += QString::fromUtf8(Helper::escapeHtml(val).c_str());
    out += "</td>";
    
    val.clear();
    e.getrprop(0, "duration", val);
    int seconds = upnpdurationtos(val);
    // Format as mm:ss
    int mins = seconds / 60;
    int secs = seconds % 60;
    char sdur[100];
    sprintf(sdur, "%02d:%02d", mins, secs);
    out += "<td class='tk_duration'>";
    out += sdur;
    out += "</td>";

    out += "</tr>";

    return out;
}


// HTML for the container display page, next after top: close head and add body tag, with some
// Javascript for webengine.
//
// We would like to use F10 for the context menu activation, because this appears to be the
// standard on MS Windows, but could not make it work, probably the desktop eats it.
static const QString init_container_pagemid = QString::fromUtf8(
    "</head><body>"
#ifdef USING_WEBENGINE
   R"(<script>
    document.addEventListener('contextmenu', saveLoc, true);
    document.addEventListener('focus', saveLoc, true);
    new QWebChannel(qt.webChannelTransport, function (channel) {
                            document.addEventListener("keydown", function(e) {
                                //console.log("EventListener:keydown: " + e.key + " " + e.shiftKey);
                                if (e.key === "ContextMenu" || (e.key === "F9" && e.shiftKey)) {
                                    channel.objects.popup_helper.onTriggered();
                                }
                            });
                        });
    </script>)"
#endif
    );

// Container display page, table header for table version.
static const QString init_container_empty_table = QString::fromUtf8(
    "<table id='entrylist'>"
    "<colgroup>"
    "<col class='coltracknumber'>"
    "<col class='coltitle'>"
    "<col class='colartist'>"
    "<col class='colalbum'>"
    "<col class='colduration'>"
    "</colgroup>"
    "</table></body></html>"
    );

// Container display page: thumbnails display area. Stays empty if we
// are in list mode and do not display album thumbnails
static const QString init_container_thumbnails_div = QString::fromUtf8(
    "<div id='thumbnailsdiv'>"
    "</div>"
    );

// Container display alphabetic jump links
QString CDBrowser::alphalinks(const std::string& initials)
{
    QString html = QString("<div class='alphalist'>");
    html += "<br clear='all'/>";
    for (unsigned int i = 0; i < initials.size(); i++) {
        QString letter(1, initials[i]);
        html += QString("<a href='http://h/a") + letter +
            QString("' tabindex='-1'>") + letter + QString("</a>");
    }

    html += QString("</div>");
    return html;
}

// Container display page: header part before the table: current
// browse path and default alphalinks
QString CDBrowser::headerHtml()
{
    QString headerdata;
    QFontMetrics metrics(font());
    headerdata += "<div id='browsepath'>\n";
    headerdata += "  <ul>\n";
    bool current_is_search = false;

    bool elide{false};
    if (m_elidePath) {
        QString path;
        for (unsigned i = 0; i < m_curpath.size(); i++) {
            QString sep("&gt;");
            if (i == 0)
                sep = "";
            path += u8s2qs(m_curpath[i].title) + sep;
        }
        // This does not work well the first time we are called (when restoring from state), because
        // the width is wrong.
        if (metrics.size(0, path).width() > width()) {
            elide = true;
        }
        LOGDEB1("CDBrowser::headerHtml: width " << width() << " header w " <<
                metrics.size(0, path).width()  << "\n");
    }

    for (unsigned i = 0; i < m_curpath.size(); i++) {
        CtPathElt& e{m_curpath[i]};
        QString sep("&gt;");
        if (!e.searchStr.empty()) {
            if (current_is_search) {
                // Indicate that searches are not nested by changing
                // the separator
                sep = "&lt;&gt;";
            }
            current_is_search = true;
        }
        if (i == 0)
            sep = "";
        std::string title;
        if (elide && e.title.length() > 5 && i != m_curpath.size()-1) {
            title = e.title.substr(0,1) + "..." + e.title.substr(e.title.length()-1, 1);
        } else {
            title = e.title;
        }
        headerdata += QString(
            "   <li class='container' objid='%3' ispl='%5'>"
            " %4 <a href='http://h/L%1'>%2</a></li>\n").
            arg(i).arg(u8s2qs(title)).arg(u8s2qs(e.objid)).arg(sep)
            .arg(e.isPlaylist);
    }

    headerdata += "  </ul>\n";
    headerdata += "</div>\n";

    headerdata += alphalinks("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
    
    return headerdata;
}

// Container display page: assemble the whole thing.
void CDBrowser::initContainerHtml(const std::string& ss)
{
    LOGDEB1("CDBrowser::initContainerHtml\n");
    QString header;
    header += "<div id='fixedheader'>\n";
    header += headerHtml();
    header += "</div>\n";
    // See comments in the CSS file
    header += "<div id='hiddenheader'>\n";
    header += headerHtml();
    header += "</div>\n";
    header += "<br clear='all'/>";

    if (!ss.empty()) {
        header += QString("<p style='padding-bottom:1em'>Search results for: ") + u8s2qs(ss);
    }
    QString html = html_top + init_container_pagemid + header + init_container_thumbnails_div +
        init_container_empty_table;
    LOGDEB1("initContainerHtml: initial content: " << qs2utf8s(html) << "\n");
    mySetHtml(html);
    LOGDEB1("CDBrowser::initContainerHtml done\n");
}
